#define  _CRT_SECURE_NO_WARNINGS
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include <windows.h>

#define MNIST_TRAINING_SET_IMAGE_FILE_NAME "data/train-images-idx3-ubyte" ///< MNIST image training file in the data folder
#define MNIST_TRAINING_SET_LABEL_FILE_NAME "data/train-labels-idx1-ubyte" ///< MNIST label training file in the data folder

#define MNIST_TESTING_SET_IMAGE_FILE_NAME "data/t10k-images-idx3-ubyte"  ///< MNIST image testing file in the data folder
#define MNIST_TESTING_SET_LABEL_FILE_NAME "data/t10k-labels-idx1-ubyte"  ///< MNIST label testing file in the data folder

#define MNIST_MAX_TRAINING_IMAGES 60000                     ///< number of images+labels in the TRAIN file/s
#define MNIST_MAX_TESTING_IMAGES 10000                      ///< number of images+labels in the TEST file/s
#define MNIST_IMG_WIDTH 28                                  ///< image width in pixel
#define MNIST_IMG_HEIGHT 28                                 ///< image height in pixel

typedef struct MNIST_ImageFileHeader MNIST_ImageFileHeader;
typedef struct MNIST_LabelFileHeader MNIST_LabelFileHeader;

typedef struct MNIST_Image MNIST_Image;
typedef uint8_t MNIST_Label;

unsigned int vram[28 * 28];

unsigned int cxClient;
unsigned int cyClient;
#define WM_TICK 1


struct MNIST_Image {
	uint8_t pixel[28 * 28];
};

//  http://yann.lecun.com/exdb/mnist/ for details

struct MNIST_ImageFileHeader
{
	uint32_t magicNumber;
	uint32_t maxImages;
	uint32_t imgWidth;
	uint32_t imgHeight;
};

void set_pixel(int x, int y, unsigned int value)
{
	if (x >= 28 || x < 0)
		return;
	if (y >= 28 || y < 0)
		return;

	vram[x + 28 * y] = value;
}

void clear_pixels()
{
	memset(vram, ~0, 28 * 28 * 4);
}

struct MNIST_LabelFileHeader
{
	uint32_t magicNumber;
	uint32_t maxImages;
};

FILE *openMNISTImageFile(char *fileName);
FILE *openMNISTLabelFile(char *fileName);
MNIST_Image getImage(FILE *imageFile);
MNIST_Label getLabel(FILE *labelFile);

MNIST_Image getImage(FILE *imageFile)
{
	MNIST_Image img;
	size_t result;

	result = fread(&img, sizeof(img), 1, imageFile);
	if (result != 1) {
		printf("\nError when reading IMAGE file! Abort!\n");
		exit(1);
	}

	return img;
}

MNIST_Label getLabel(FILE *labelFile)
{
	MNIST_Label lbl;
	size_t result;
	result = fread(&lbl, sizeof(lbl), 1, labelFile);
	if (result != 1)
	{
		printf("\nError when reading LABEL file! Abort!\n");
		exit(1);
	}

	return lbl;
}

uint32_t flipBytes(uint32_t n)
{
	uint32_t b0, b1, b2, b3;

	b0 = (n & 0x000000ff) << 24u;
	b1 = (n & 0x0000ff00) << 8u;
	b2 = (n & 0x00ff0000) >> 8u;
	b3 = (n & 0xff000000) >> 24u;

	return (b0 | b1 | b2 | b3);

}

void readImageFileHeader(FILE *imageFile, MNIST_ImageFileHeader *ifh)
{
	ifh->magicNumber = 0;
	ifh->maxImages = 0;
	ifh->imgWidth = 0;
	ifh->imgHeight = 0;

	fread(&ifh->magicNumber, 4, 1, imageFile);
	ifh->magicNumber = flipBytes(ifh->magicNumber);

	fread(&ifh->maxImages, 4, 1, imageFile);
	ifh->maxImages = flipBytes(ifh->maxImages);

	fread(&ifh->imgWidth, 4, 1, imageFile);
	ifh->imgWidth = flipBytes(ifh->imgWidth);

	fread(&ifh->imgHeight, 4, 1, imageFile);
	ifh->imgHeight = flipBytes(ifh->imgHeight);
}

FILE *openMNISTImageFile(char *fileName)
{
	FILE *imageFile;
	imageFile = fopen(fileName, "rb");
	if (imageFile == NULL)
	{
		return NULL;
	}

	MNIST_ImageFileHeader imageFileHeader;
	readImageFileHeader(imageFile, &imageFileHeader);

	return imageFile;
}

void readLabelFileHeader(FILE *imageFile, MNIST_LabelFileHeader *lfh)
{
	lfh->magicNumber = 0;
	lfh->maxImages = 0;

	fread(&lfh->magicNumber, 4, 1, imageFile);
	lfh->magicNumber = flipBytes(lfh->magicNumber);

	fread(&lfh->maxImages, 4, 1, imageFile);
	lfh->maxImages = flipBytes(lfh->maxImages);

}

FILE *openMNISTLabelFile(char *fileName)
{
	FILE *labelFile;
	labelFile = fopen(fileName, "rb");
	if (labelFile == NULL)
	{
		return NULL;
	}

	MNIST_LabelFileHeader labelFileHeader;
	readLabelFileHeader(labelFile, &labelFileHeader);

	return labelFile;
}

#define NUMBER_OF_INPUT_CELLS 784   ///< use 28*28 input cells (= number of pixels per MNIST image)
#define NUMBER_OF_OUTPUT_CELLS 10   ///< use 10 output cells to model 10 digits (0-9)

#define LEARNING_RATE 0.05          ///< Incremental increase for changing connection weights

typedef struct
{
	double input[NUMBER_OF_INPUT_CELLS];
	double weight[NUMBER_OF_INPUT_CELLS];
	double output;
	double bias;
} Cell;

typedef struct
{
	Cell cell[NUMBER_OF_OUTPUT_CELLS];
} Layer;

typedef struct
{
	int val[NUMBER_OF_OUTPUT_CELLS];
} Vector;

Vector getTargetOutput(int targetIndex);
void initLayer(Layer *l);
int getLayerPrediction(Layer *l);
void setCellInput(Cell *c, MNIST_Image *img);
void calcCellOutput(Cell *c);
double getCellError(Cell *c, int target);
void updateCellWeights(Cell *c, double err);
void trainCell(Cell *c, MNIST_Image *img, int target);
void testCell(Cell *c, MNIST_Image *img, int target);

Vector getTargetOutput(int targetIndex)
{
	Vector v;
	for (int i = 0; i<NUMBER_OF_OUTPUT_CELLS; i++) {
		v.val[i] = (i == targetIndex) ? 1 : 0;
	}
	return v;
}

int getLayerPrediction(Layer *l)
{

	double maxOut = 0;
	int maxInd = 0;

	for (int i = 0; i<NUMBER_OF_OUTPUT_CELLS; i++) {

		if (l->cell[i].output > maxOut) {
			maxOut = l->cell[i].output;
			maxInd = i;
		}
	}

	return maxInd;

}

void setCellInput(Cell *c, MNIST_Image *img)
{
	for (int i = 0; i<NUMBER_OF_INPUT_CELLS; i++) {
		c->input[i] = img->pixel[i] ? 1 : 0;
	}
}


void calcCellOutput(Cell *c)
{
	c->output = 0;

	for (int i = 0; i<NUMBER_OF_INPUT_CELLS; i++) {
		c->output += c->input[i] * c->weight[i];
	}

	c->output /= NUMBER_OF_INPUT_CELLS;             // normalize output (0-1)
}

double getCellError(Cell *c, int target)
{
	double err = target - c->output;

	return err;
}

void updateCellWeights(Cell *c, double err)
{

	for (int i = 0; i<NUMBER_OF_INPUT_CELLS; i++) {
		c->weight[i] += LEARNING_RATE * c->input[i] * err;
	}
}

void trainCell(Cell *c, MNIST_Image *img, int target)
{

	setCellInput(c, img);
	calcCellOutput(c);

	// learning (by updating the weights)
	double err = getCellError(c, target);
	updateCellWeights(c, err);
}

void testCell(Cell *c, MNIST_Image *img, int target)
{
	setCellInput(c, img);
	calcCellOutput(c);
}



void trainLayer(Layer *l)
{
    // open MNIST files
    FILE *imageFile, *labelFile;
    imageFile = openMNISTImageFile(MNIST_TRAINING_SET_IMAGE_FILE_NAME);
    labelFile = openMNISTLabelFile(MNIST_TRAINING_SET_LABEL_FILE_NAME);
    
    int errCount = 0;
    
    // Loop through all images in the file
    for (int imgCount=0; imgCount<MNIST_MAX_TRAINING_IMAGES; imgCount++)
	{                
        // Reading next image and corresponding label
        MNIST_Image img = getImage(imageFile);
        MNIST_Label lbl = getLabel(labelFile);
        
        // set target ouput of the number displayed in the current image (=label) to 1, all others to 0
        Vector targetOutput;
        targetOutput = getTargetOutput(lbl);
             
        // loop through all output cells for the given image
        for (int i=0; i < NUMBER_OF_OUTPUT_CELLS; i++)
		{
            trainCell(&l->cell[i], &img, targetOutput.val[i]);
        }
        
        int predictedNum = getLayerPrediction(l);
        if (predictedNum!=lbl)
			errCount++;
    }
    
    // Close files
    fclose(imageFile);
    fclose(labelFile);

}


int testLayer(Layer *l)
{    
    int errCount = 0;
	MNIST_Image img;

	memset(img.pixel, 0, 28 * 28);
	for (int y = 0; y < 28; y++)
	{
		for (int x = 0; x < 28; x++)
		{
			img.pixel[x + y * 28] = 0xFF & ~vram[x + y * 28];
		}
	}
        
    // set target ouput of the number displayed in the current image (=label) to 1, all others to 0
    Vector targetOutput;
	MNIST_Label lbl;

	lbl = '0';
    targetOutput = getTargetOutput(lbl);

    // loop through all output cells for the given image
    for (int i=0; i < NUMBER_OF_OUTPUT_CELLS; i++)
	{
        testCell(&l->cell[i], &img, targetOutput.val[i]);
    }
        
    int predictedNum = getLayerPrediction(l);
        
	return predictedNum;
}


LRESULT CALLBACK WndProc(HWND, UINT, WPARAM, LPARAM);

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance,
	PSTR szCmdLine, int iCmdShow)
{
	static TCHAR szAppName[] = TEXT("NeuralNetwork");
	HWND         hwnd;
	MSG          msg;
	WNDCLASS     wndclass;

	wndclass.style = CS_HREDRAW | CS_VREDRAW;
	wndclass.lpfnWndProc = WndProc;
	wndclass.cbClsExtra = 0;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = hInstance;
	wndclass.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	wndclass.hCursor = LoadCursor(NULL, IDC_ARROW);
	wndclass.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wndclass.lpszMenuName = NULL;
	wndclass.lpszClassName = szAppName;

	if (!RegisterClass(&wndclass))
	{
		MessageBox(NULL, TEXT("Program requires Windows NT!"),
			szAppName, MB_ICONERROR);
		return 0;
	}

	hwnd = CreateWindow(szAppName, TEXT("Neural Network Number Recognition"),
		WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT, 200,
		CW_USEDEFAULT, 250,
		NULL, NULL, hInstance, NULL);

	ShowWindow(hwnd, iCmdShow);
	UpdateWindow(hwnd);

	while (GetMessage(&msg, NULL, 0, 0))
	{
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}
	return msg.wParam;
}


void initLayer(Layer *l)
{
	for (int o = 0; o < NUMBER_OF_OUTPUT_CELLS; o++)
	{

		for (int i = 0; i < NUMBER_OF_INPUT_CELLS; i++)
		{
			l->cell[o].input[i] = 0;
			l->cell[o].weight[i] = rand() / (double)(RAND_MAX);
		}

		l->cell[o].output = 0;
		l->cell[o].bias = 0;
	}
}

void draw_pixels(HDC hdc, HDC hdcMem, int width, int height)
{
	HBITMAP hBitmap, hOldBitmap;

	hBitmap = CreateCompatibleBitmap(hdc, width, height);
	SetBitmapBits(hBitmap, sizeof(int) * width * height, vram);
	hdcMem = CreateCompatibleDC(hdc);
	hOldBitmap = (HBITMAP)SelectObject(hdcMem, hBitmap);

	// This scaling is a little strange because Stretch maintains aspect ratios
	StretchBlt(hdc, 0, 0, cxClient, cyClient, hdcMem, 0, 0, width, height, SRCCOPY);
	SelectObject(hdcMem, hOldBitmap);
	DeleteDC(hdcMem);
	DeleteObject(hBitmap);
}



void load_random()
{
	int sample = rand() % MNIST_MAX_TRAINING_IMAGES;

	FILE *imageFile, *labelFile;
	imageFile = openMNISTImageFile(MNIST_TRAINING_SET_IMAGE_FILE_NAME);
	labelFile = openMNISTLabelFile(MNIST_TRAINING_SET_LABEL_FILE_NAME);

	for (int imgCount = 0; imgCount < sample; imgCount++)
	{
		// Reading next image and corresponding label
		MNIST_Image img = getImage(imageFile);
		MNIST_Label lbl = getLabel(labelFile);

		for (int y = 0; y < 28; y++)
		{
			for (int x = 0; x < 28; x++)
			{
				vram[x + y * 28] = RGB(~img.pixel[x + y * 28], ~img.pixel[x + y * 28], ~img.pixel[x + y * 28]);
			}
		}

	}

	// Close files
	fclose(imageFile);
	fclose(labelFile);

}


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	static Layer outputLayer;
	static RECT	rect = { 0, 0, 1, 1 };
	static HDC	hdcMem;
	PAINTSTRUCT ps;
	HDC hdc;
	static int	old_style = 0;
	static int	new_style = WS_CHILD | WS_VISIBLE;
	static int	xres, yres;
	static int guess = -1;



	switch (message)
	{
	case WM_TIMER:
	{
		int width = cxClient;
		int height = cyClient;

		InvalidateRect(hwnd, &rect, FALSE);
		return 0;
	}
	case WM_CREATE:
	{
		initLayer(&outputLayer);

//		printf("Training Neural Network\r\n");
		trainLayer(&outputLayer);
//		printf("Done\r\n");

		memset(vram, ~0, 28 * 28 * 4);
		xres = 28;
		yres = 28;

		SetWindowPos(hwnd, 0, 200, 200, 200, 250, 0);
		SetTimer(hwnd, WM_TICK, 16, NULL);
		return 0;
	}
	case WM_SIZE:
		cxClient = LOWORD(lParam);
		cyClient = HIWORD(lParam);
		rect.right = cxClient;
		rect.bottom = cyClient;
		return 0;

	case WM_KEYDOWN:
		switch (wParam)
		{
		case VK_SPACE:
			clear_pixels();
			break;
		case VK_RETURN:
			guess = testLayer(&outputLayer);			
			break;
		case VK_TAB:
			load_random();
			break;

		}
		return 0;

	case WM_MOUSEMOVE:

		if (wParam & MK_LBUTTON)
		{
			float x = (float)LOWORD(lParam) / cxClient;
			float y = (float)HIWORD(lParam) / cyClient;

			set_pixel((int)(x * xres),			(int)(y * yres), 0); // center
			set_pixel((int)(x * xres + 1),		(int)(y * yres), 0); // right
			set_pixel((int)(x * xres - 1),		(int)(y * yres), 0);   // left
			set_pixel((int)(x * xres),			(int)(y * yres + 1), 0); // up
			set_pixel((int)(x * xres),			(int)(y * yres - 1), 0); // down
			set_pixel((int)(x * xres + 1),		(int)(y * yres + 1), 0); // up right
			set_pixel((int)(x * xres - 1),		(int)(y * yres - 1), 0); // down left
			set_pixel((int)(x * xres - 1),		(int)(y * yres + 1), 0); // up left
			set_pixel((int)(x * xres + 1),		(int)(y * yres - 1), 0); // down right
		}

		return 0;

	case WM_PAINT:
	{
		char guess_str[80];

		sprintf(guess_str, "Predicted: %d", guess);
		hdc = BeginPaint(hwnd, &ps);
		draw_pixels(hdc, hdcMem, xres, yres);
		DrawText(hdc, guess_str, -1, &rect, 0);
		EndPaint(hwnd, &ps);
		return 0;
	}
	case WM_DESTROY:
		PostQuitMessage(0);
		return 0;
	}
	return DefWindowProc(hwnd, message, wParam, lParam);
}

